# yapf: disable
# isort: skip_file
# ruff: noqa

# Standard imports.
from __future__ import annotations

import asyncio
import google.protobuf.descriptor
import google.protobuf.json_format
import google.protobuf.message
import grpc
import json
import logging
import sys
import traceback
import uuid
import resemble
import respect.logging  # type: ignore[import]
import typing
from abc import abstractmethod
from backoff import Backoff
from datetime import datetime, timedelta
from google.protobuf import empty_pb2
from google.rpc import status_pb2
from grpc_status import rpc_status
from resemble.aio.call import InvalidActorIdError, MixedContextsError
from resemble.aio.contexts import (
    Context,
    EffectValidationRetry,
    ReaderContext,
    TransactionContext,
    WorkflowContext,
    WriterContext,
    EffectValidation,
)
from resemble.aio.headers import Headers
from resemble.aio.idempotency import IdempotencyManager, Idempotency
from resemble.aio.internals.channel_manager import _ChannelManager
from resemble.aio.internals.middleware import (
    maybe_run_function_twice_to_validate_effects,
    maybe_run_method_twice_to_validate_effects,
    Middleware,
)
from resemble.aio.internals.tasks_cache import TasksCache
from resemble.aio.internals.tasks_dispatcher import TasksDispatcher
from resemble.aio.headers import Headers
from resemble.aio.servicers import Servicer, Serviceable
from resemble.aio.state_managers import Effects, StateManager
from resemble.aio.stubs import Stub
from resemble.aio.tasks import TaskEffect
from resemble.aio.types import (
    ActorId,
    ApplicationId,
    GrpcMetadata,
    ServiceName,
    assert_type,
)
from resemble.aio.workflows import Workflow
from resemble.settings import MAX_ACTOR_ID_LENGTH
from resemble.v1alpha1 import tasks_pb2, tasks_pb2_grpc
from typing import (
    cast,
    Any,
    AsyncIterable,
    AsyncIterator,
    Awaitable,
    Callable,
    Generator,
    NoReturn,
    Optional,
    TypeAlias,
    Union,
)

# User defined or referenced imports.
import google.protobuf.timestamp_pb2
import resemble.integrations.mailgun.v1.mailgun_pb2
import resemble.integrations.mailgun.v1.mailgun_pb2_grpc
import resemble.v1alpha1.options_pb2
import resemble.v1alpha1.tasks_pb2
# Additionally re-export all messages and enums from the pb2 module.
from resemble.integrations.mailgun.v1.mailgun_pb2 import (
    MessageState,
    SendRequest,
    SendResponse,
    SendTaskRequest,
    SendTaskResponse,
)

logger = respect.logging.get_logger(__name__)


def MakeLegacyGrpcServiceable(
    # A legacy gRPC servicer type can't be more specific than `type`,
    # because legacy gRPC servicers (as generated by the gRPC `protoc`
    # plugin) do not share any common base class other than `object`.
    servicer_type: type
) -> Serviceable:
    raise ValueError(f"Unknown legacy gRPC servicer type '{servicer_type}'")


# For internal calls, we can use a magic token to bypass token verification and
# authorization checks. The token provides no auth information (e.g.,
# `context.auth is None`).
__internal_magic_token__: str = f'internal-{str(uuid.uuid4())}'

class MessageServicerMiddleware(Middleware):

    def __init__(
        self,
        *,
        servicer: MessageServicer,
        application_id: ApplicationId,
        state_manager: StateManager,
        channel_manager: _ChannelManager,
        tasks_cache: TasksCache,
        token_verifier: Optional[resemble.aio.auth.token_verifiers.TokenVerifier],
        effect_validation: EffectValidation,
        ready: asyncio.Event,
    ):
        super().__init__(
            application_id=application_id,
            service_name='resemble.integrations.mailgun.v1.Message',
            channel_manager=channel_manager,
            effect_validation=effect_validation,
        )

        self._servicer = servicer
        self._state_manager = state_manager
        self.tasks_dispatcher = TasksDispatcher(
            application_id=application_id,
            dispatch=self.dispatch,
            tasks_cache=tasks_cache,
            ready=ready,
        )

        # Store the type of each method's request so that stored requests can be
        # deserialized into the correct type.
        self.request_type_by_method_name: dict[str, type[google.protobuf.message.Message]] = {
            'Send': resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest,
            'SendTask': resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest,
        }

        # Create authorizer.
        self._authorizer: Optional[resemble.aio.auth.authorizers.Authorizer] = servicer.authorizer()

        # Create token verifier.
        self._token_verifier: Optional[resemble.aio.auth.token_verifiers.TokenVerifier] = (
            servicer.token_verifier() or token_verifier
        )

        if self._token_verifier is not None and self._authorizer is None:
            raise RuntimeError(
                f"'TokenVerifier' provided but no 'Authorizer' returned for "
                f"'resemble.integrations.mailgun.v1.Message'"
            )

    def add_to_server(self, server: grpc.aio.Server) -> None:
        resemble.integrations.mailgun.v1.mailgun_pb2_grpc.add_MessageServicer_to_server(
            self, server
        )

    async def inspect(self, actor_id: ActorId) -> AsyncIterator[google.protobuf.message.Message]:
        """Implementation of `Middleware.inspect()`."""
        context = self.create_context(
            headers=Headers(
                application_id=self.application_id,
                service_name=self.service_name,
                actor_id=actor_id,
            ),
            context_type=ReaderContext,
        )

        async with self._state_manager.streaming_reader_idempotency_key(
            context, self._servicer.__state_type__
        ) as states:
            async for (state, idempotency_key) in states:
                yield state

    async def react_query(
        self,
        grpc_context: grpc.aio.ServicerContext,
        headers: Headers,
        method: str,
        request_bytes: bytes,
    ) -> AsyncIterator[tuple[Optional[google.protobuf.message.Message], list[uuid.UUID]]]:
        """Returns the response of calling 'method' given a message
        deserialized from the provided 'request_bytes' for each state
        update that creates a different response.

        NOTE: only unary reader methods are supported."""
        # Need to define these up here since we can only do that once.
        last_response: Optional[google.protobuf.message.Message] = None
        aggregated_idempotency_keys: list[uuid.UUID] = []
        if method == 'Send':
            # Invariant here is that users should not have called this
            # directly but only through code generated React
            # components which should not have been generated except
            # for valid method candidates.
            logger.warning(
                "Got a React query request with an invalid method name: "
                "Method '{method}' is invalid for servicer Message."
                "\n"
                "Do you have a browser tab open for an older version "
                "of this application, or for a different application all together?"
            )
            raise resemble.aio.aborted.SystemAborted(
                resemble.v1alpha1.errors_pb2.InvalidMethod(),
                message=
                    f"Method '{method}' is invalid"
            )
            yield  # Necessary for type checking.
        elif method == 'SendTask':
            # Invariant here is that users should not have called this
            # directly but only through code generated React
            # components which should not have been generated except
            # for valid method candidates.
            logger.warning(
                "Got a React query request with an invalid method name: "
                "Method '{method}' is invalid for servicer Message."
                "\n"
                "Do you have a browser tab open for an older version "
                "of this application, or for a different application all together?"
            )
            raise resemble.aio.aborted.SystemAborted(
                resemble.v1alpha1.errors_pb2.InvalidMethod(),
                message=
                    f"Method '{method}' is invalid"
            )
            yield  # Necessary for type checking.
        else:
            logger.warning(
                "Got a React query request with an invalid method name: "
                "Method '{method}' is invalid for servicer Message."
                "\n"
                "Do you have a browser tab open for an older version "
                "of this application, or for a different application all together?"
            )
            raise resemble.aio.aborted.SystemAborted(
                resemble.v1alpha1.errors_pb2.InvalidMethod(),
                message=
                    f"Method '{method}' not found"
            )
            yield  # Unreachable but necessary for mypy.

    @maybe_run_method_twice_to_validate_effects
    async def react_mutate(
        self,
        validating_effects: bool,
        headers: Headers,
        method: str,
        request_bytes: bytes,
    ) -> google.protobuf.message.Message:
        """Returns the response of calling 'method' given a message
        deserialized from the provided 'request_bytes'."""
        if method == 'Send':
            request = resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest()
            request.ParseFromString(request_bytes)

            # NOTE: we automatically retry mutations that come through
            # React when we get a `grpc.StatusCode.UNAVAILABLE` to
            # match the retry logic we do in the React code generated
            # to handle lack/loss of connectivity.
            #
            # TODO(benh): revisit this decision if we ever see reason
            # to call `react_mutate()` from any place other than where
            # we're executing React (e.g., browser, next.js server
            # component, etc).
            backoff = Backoff()
            while True:
                try:
                    return await self._Send(
                        request,
                        self.create_context(
                            headers=headers,
                            context_type=WriterContext,
                        ),
                        validating_effects=validating_effects,
                    )
                except resemble.aio.aborted.Aborted as aborted:
                    if aborted.code == grpc.StatusCode.UNAVAILABLE:
                        await backoff()
                        continue
                    raise
        elif method == 'SendTask':
            request = resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest()
            request.ParseFromString(request_bytes)

            # NOTE: we automatically retry mutations that come through
            # React when we get a `grpc.StatusCode.UNAVAILABLE` to
            # match the retry logic we do in the React code generated
            # to handle lack/loss of connectivity.
            #
            # TODO(benh): revisit this decision if we ever see reason
            # to call `react_mutate()` from any place other than where
            # we're executing React (e.g., browser, next.js server
            # component, etc).
            backoff = Backoff()
            while True:
                try:
                    return await self._SendTask(
                        request,
                        self.create_context(
                            headers=headers,
                            context_type=WriterContext,
                        ),
                        validating_effects=validating_effects,
                    )
                except resemble.aio.aborted.Aborted as aborted:
                    if aborted.code == grpc.StatusCode.UNAVAILABLE:
                        await backoff()
                        continue
                    raise
        else:
            logger.warning(
                "Got a react mutate request with an invalid method name: "
                "Method '{method}' is invalid for servicer Message."
                "\n"
                "Do you have an old browser tab still open for an older version "
                "of this application, or a different application all together?"
            )
            raise resemble.aio.aborted.SystemAborted(
                resemble.v1alpha1.errors_pb2.InvalidMethod(),
                message=
                    f"Method '{method}' not found"
            )

    async def dispatch(
        self,
        task: TaskEffect,
        *,
        only_validate: bool = False,
    ) -> google.protobuf.message.Message | resemble.aio.tasks.Loop:
        """Dispatches the tasks to execute unless 'only_validate' is set to
        true, in which case just ensures that the task actually exists.
        Note that this function will be called *by* tasks_dispatcher; it will
        not itself call into tasks_dispatcher."""

        if 'Send' == task.method_name:
            if only_validate:
                # TODO(benh): validate 'task.request' is correct type.
                return resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse()

            # Use an inline method to create a new scope, so that we can use
            # variable names like `context` and `effects` in multiple branches
            # in this code (notably when there are multiple task types) without
            # hitting a mypy error that the variable's type is not consistent.
            @maybe_run_function_twice_to_validate_effects
            async def run_Send(
                validating_effects: bool,
                context: WorkflowContext,
            ):
                async with self._state_manager.task_workflow(
                    context,
                ) as complete:
                    response = await (MessageWorkflowStub(
                        context_or_workflow=context,
                        actor_id=context.actor_id,
                    ).Send(
                        cast(resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest, task.request),
                        bearer_token=__internal_magic_token__,
                        idempotency=Idempotency(
                            alias=f'Task {task.task_id.task_uuid}',
                        ),
                    ))

                    await complete(task, response)
                    return response




            return await run_Send(
                self.create_context(
                    headers=Headers(
                        application_id=self.application_id,
                        service_name=self.service_name,
                        actor_id=task.task_id.actor_id,
                    ),
                    context_type=WorkflowContext,
                    task=task,
                )
            )
        elif 'SendTask' == task.method_name:
            if only_validate:
                # TODO(benh): validate 'task.request' is correct type.
                return resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse()

            # Use an inline method to create a new scope, so that we can use
            # variable names like `context` and `effects` in multiple branches
            # in this code (notably when there are multiple task types) without
            # hitting a mypy error that the variable's type is not consistent.
            @maybe_run_function_twice_to_validate_effects
            async def run_SendTask(
                validating_effects: bool,
                context: WorkflowContext,
            ):
                async with self._state_manager.task_workflow(
                    context,
                ) as complete:
                    response = await (MessageWorkflowStub(
                        context_or_workflow=context,
                        actor_id=context.actor_id,
                    ).SendTask(
                        cast(resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest, task.request),
                        bearer_token=__internal_magic_token__,
                        idempotency=Idempotency(
                            alias=f'Task {task.task_id.task_uuid}',
                        ),
                    ))

                    await complete(task, response)
                    return response




            return await run_SendTask(
                self.create_context(
                    headers=Headers(
                        application_id=self.application_id,
                        service_name=self.service_name,
                        actor_id=task.task_id.actor_id,
                    ),
                    context_type=WorkflowContext,
                    task=task,
                )
            )

        # There are no tasks for this service.
        start_or_validate = "start" if not only_validate else "validate"
        raise RuntimeError(
            f"Attempted to {start_or_validate} task '{task.method_name}' "
            f"on 'Message' which does not exist"
        )

    # Message specific methods:
    async def __Send(
        self,
        context: WriterContext,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest,
        *,
        validating_effects: bool,
    ) -> Message.SendEffects:
        try:
            response = (
                await self._servicer.Send(
                    context=context,
                    request=request
                )
            )
            self.maybe_raise_effect_validation_retry(
                logger=logger,
                idempotency_manager=context,
                method_name='Message.Send',
                validating_effects=validating_effects,
                is_non_root_in_transaction=(
                    not issubclass(type(context), TransactionContext) and
                    context.transaction_id is not None
                ),
            )
            return response
        except EffectValidationRetry:
            # Doing effect validation, just let this propagate.
            raise
        except resemble.aio.aborted.Aborted as aborted:
            # Log any _unhandled_ abort stack traces to make it
            # easier for debugging.
            #
            # NOTE: we don't log if we're a task as it will be logged
            # in `resemble/aio/internals/tasks_dispatcher.py` instead.
            aborted_type: Optional[type] = None
            aborted_type = Message.SendAborted
            if isinstance(aborted, resemble.aio.aborted.SystemAborted):
                if context.task is None:
                    logger.warning(
                        f"Unhandled (in 'resemble.integrations.mailgun.v1.Message.Send') {aborted}; propagating as 'Unknown'\n" +
                        ''.join(traceback.format_exception(aborted))
                    )
                raise resemble.aio.aborted.SystemAborted(
                    resemble.v1alpha1.errors_pb2.Unknown(),
                    # TODO(benh): consider whether or not we want to
                    # include the 'package.service.method' which may
                    # get concatenated together forming a kind of
                    # "stack trace"; while it's super helpful for
                    # debugging, it does expose implementation
                    # information.
                    message=f"unhandled (in 'resemble.integrations.mailgun.v1.Message.Send') {aborted}"
                )
            else:
                if (
                    aborted_type is not None and
                    not isinstance(aborted, aborted_type) and
                    aborted_type.is_declared_error(aborted.error)
                ):
                    if context.task is None:
                        logger.warning(
                            f"Propagating unhandled (in 'resemble.integrations.mailgun.v1.Message.Send') {aborted}"
                        )
                elif (
                    aborted_type is None or
                    not isinstance(aborted, aborted_type)
                ):
                    if context.task is None:
                        logger.warning(
                            f"Unhandled (in 'resemble.integrations.mailgun.v1.Message.Send') {aborted}; propagating as 'Unknown'\n" +
                            ''.join(traceback.format_exception(aborted))
                        )
                    raise resemble.aio.aborted.SystemAborted(
                        resemble.v1alpha1.errors_pb2.Unknown(),
                        # TODO(benh): consider whether or not we want to
                        # include the 'package.service.method' which may
                        # get concatenated together forming a kind of
                        # "stack trace"; while it's super helpful for
                        # debugging, it does expose implementation
                        # information.
                        message=f"unhandled (in 'resemble.integrations.mailgun.v1.Message.Send') {aborted}"
                    )

            raise
        except asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except BaseException as exception:
            if context.task is None:
                logger.warning(
                    f"Unhandled (in 'resemble.integrations.mailgun.v1.Message.Send') {type(exception).__name__}: {exception}; propagating as 'Unknown'\n" +
                    ''.join(traceback.format_exception(exception))
                )
            raise resemble.aio.aborted.SystemAborted(
                resemble.v1alpha1.errors_pb2.Unknown(),
                # TODO(benh): consider whether or not we want to
                # include the 'package.service.method' which may
                # get concatenated together forming a kind of
                # "stack trace"; while it's super helpful for
                # debugging, it does expose implementation
                # information.
                message=f"unhandled (in 'resemble.integrations.mailgun.v1.Message.Send') {type(exception).__name__}: {exception}"
            )

    async def _Send(
        self,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest,
        context: WriterContext,
        *,
        validating_effects: bool,
        grpc_context: Optional[grpc.aio.ServicerContext] = None,
    ) -> resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse:
        # Try to verify the token if a token verifier exists.
        context.auth = await self._maybe_verify_token(context._headers)

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        idempotent_mutation = self._state_manager.check_for_idempotent_mutation(
            context
        )

        if idempotent_mutation is not None:
            response = resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse()
            response.ParseFromString(idempotent_mutation.response)
            return response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__service_name__, context.actor_id
                )
            # TODO: this loads a `state` object even when we know we're calling a constructor,
            # in which case (by definition) there is no state. Can we (and is it worth
            # the effort to) change this call to avoid that overhead in that case?
            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                transaction=transaction,
                from_constructor=True,
                requires_constructor=True
            ) as (state, writer):

                await self._maybe_authorize(
                    method_name=
                    'resemble.integrations.mailgun.v1.Message.Send',
                    headers=context._headers,
                    auth=context.auth,
                    request=request,
                )

                effects = await self.__Send(
                    context,
                    request,
                    validating_effects=validating_effects,
                )

                if effects.tasks is not None:
                    # NOTE: we validate tasks added as part of a
                    # transaction when we prepare.
                    if transaction is None:
                        await self.tasks_dispatcher.validate(
                            effects.tasks
                        )

                await writer.complete(effects)

                if effects.tasks is not None:
                    if transaction is None:
                        self.tasks_dispatcher.dispatch(effects.tasks)
                    else:
                        assert all(
                            task.task_id.service == transaction.service
                            for task in effects.tasks
                        ), 'Task service does not match transaction service'
                        transaction.tasks.extend(effects.tasks)

                # TODO: We need a single `Effects` superclass for all methods, so we
                # would need to make it "partially" generic (with per-method subclasses
                # filling out the rest of the generic parameters) in order to fix this.
                return effects.response  # type: ignore[return-value]

    async def _schedule_Send(
        self,
        *,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest,
        headers: Headers,
        grpc_context: grpc.aio.ServicerContext,
    ) -> tuple[WriterContext, resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse]:
        context: WriterContext = self.create_context(
            headers=headers,
            context_type=WriterContext,
        )
        response = resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse()

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        idempotent_mutation = self._state_manager.check_for_idempotent_mutation(
            context
        )

        if idempotent_mutation is not None:
            response.ParseFromString(idempotent_mutation.response)

            # We should have only scheduled a single task!
            assert len(idempotent_mutation.task_ids) == 1
            assert grpc_context is not None
            grpc_context.set_trailing_metadata(
                grpc_context.trailing_metadata() +
                (
                    (
                        resemble.aio.headers.TASK_ID_UUID,
                        str(uuid.UUID(bytes=idempotent_mutation.task_ids[0].task_uuid))
                    ),
                )
            )

            return context, response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__service_name__, context.actor_id
                )
            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                transaction=transaction,
                from_constructor=True,
                requires_constructor=True
            ) as (state, writer):

                # Try to verify the token if a token verifier exists.
                context.auth = await self._maybe_verify_token(headers)

                # Check if we are allowed schedule the task.
                await self._maybe_authorize(
                    method_name=
                    'resemble.integrations.mailgun.v1.Message.Send',
                    headers=context._headers,
                    auth=context.auth,
                    request=request,
                )

                task = MessageServicerTasksStub(
                    context=context,
                    actor_id=context.actor_id,
                ).Send(
                    request,
                    schedule=context._headers.task_schedule,
                )

                effects = Effects(
                    response=response,
                    state=state,
                    tasks=[task],
                )

                assert effects.tasks is not None

                # NOTE: we validate tasks added as part of a
                # transaction when we prepare.
                if transaction is None:
                    await self.tasks_dispatcher.validate(effects.tasks)

                await writer.complete(effects)

                if transaction is None:
                    self.tasks_dispatcher.dispatch(effects.tasks)
                else:
                    assert task.task_id.service == transaction.service
                    transaction.tasks.extend(effects.tasks)

                assert grpc_context is not None

                grpc_context.set_trailing_metadata(
                    grpc_context.trailing_metadata() +
                    (
                        (
                            resemble.aio.headers.TASK_ID_UUID,
                            str(uuid.UUID(bytes=task.task_id.task_uuid))
                        ),
                    )
                )

                return context, response

        return context, response


    @maybe_run_method_twice_to_validate_effects
    async def Send(
        self,
        validating_effects: bool,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest,
        grpc_context: grpc.aio.ServicerContext,
    ) -> resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse:
        context: Optional[Context] = None
        try:
            headers = Headers.from_grpc_context(grpc_context)
            if headers.task_schedule is not None:
                context, response = await self._schedule_Send(
                    headers=headers,
                    request=request,
                    grpc_context=grpc_context,
                )
                return response

            context = self.create_context(
                headers=headers,
                context_type=WriterContext,
            )
            assert context is not None

            return await self._Send(
                request,
                context,
                validating_effects=validating_effects,
                grpc_context=grpc_context,
            )
        except EffectValidationRetry:
            # Doing effect validation, just let this propagate.
            raise
        except resemble.aio.aborted.Aborted as aborted:
            await grpc_context.abort_with_status(
                rpc_status.to_status(aborted.to_status())
            )
            raise  # Unreachable but necessary for mypy.
        except asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except:
            # Print the exception stack trace for easier debugging. Note
            # that we don't include the stack trace in an error message
            # for the same reason that gRPC doesn't do so by default,
            # see https://github.com/grpc/grpc/issues/14897, but since this
            # should only get logged on the server side it is safe.
            logger.warning(
                'Unhandled exception\n' +
                ''.join(traceback.format_exc())
            )

            # Re-raise the exception for gRPC to handle!
            raise
        finally:
            if context is not None and context.transaction_id is not None:
                # Propagate transaction participants.
                #
                # NOTE: we'll want to propagate participants if we're
                # in a transaction method once we support nested
                # transactions, but for now we only need to propagate
                # participants when we're not in a transaction method.
                if not isinstance(context, TransactionContext):
                    grpc_context.set_trailing_metadata(
                        grpc_context.trailing_metadata() +
                        context.participants.to_grpc_metadata()
                    )

    async def __SendTask(
        self,
        context: WriterContext,
        state: resemble.integrations.mailgun.v1.mailgun_pb2.MessageState,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest,
        *,
        validating_effects: bool,
    ) -> Message.SendTaskEffects:
        try:
            response = (
                await self._servicer.SendTask(
                    context=context,
                    state=state,
                    request=request
                )
            )
            self.maybe_raise_effect_validation_retry(
                logger=logger,
                idempotency_manager=context,
                method_name='Message.SendTask',
                validating_effects=validating_effects,
                is_non_root_in_transaction=(
                    not issubclass(type(context), TransactionContext) and
                    context.transaction_id is not None
                ),
            )
            return response
        except EffectValidationRetry:
            # Doing effect validation, just let this propagate.
            raise
        except resemble.aio.aborted.Aborted as aborted:
            # Log any _unhandled_ abort stack traces to make it
            # easier for debugging.
            #
            # NOTE: we don't log if we're a task as it will be logged
            # in `resemble/aio/internals/tasks_dispatcher.py` instead.
            aborted_type: Optional[type] = None
            aborted_type = Message.SendTaskAborted
            if isinstance(aborted, resemble.aio.aborted.SystemAborted):
                if context.task is None:
                    logger.warning(
                        f"Unhandled (in 'resemble.integrations.mailgun.v1.Message.SendTask') {aborted}; propagating as 'Unknown'\n" +
                        ''.join(traceback.format_exception(aborted))
                    )
                raise resemble.aio.aborted.SystemAborted(
                    resemble.v1alpha1.errors_pb2.Unknown(),
                    # TODO(benh): consider whether or not we want to
                    # include the 'package.service.method' which may
                    # get concatenated together forming a kind of
                    # "stack trace"; while it's super helpful for
                    # debugging, it does expose implementation
                    # information.
                    message=f"unhandled (in 'resemble.integrations.mailgun.v1.Message.SendTask') {aborted}"
                )
            else:
                if (
                    aborted_type is not None and
                    not isinstance(aborted, aborted_type) and
                    aborted_type.is_declared_error(aborted.error)
                ):
                    if context.task is None:
                        logger.warning(
                            f"Propagating unhandled (in 'resemble.integrations.mailgun.v1.Message.SendTask') {aborted}"
                        )
                elif (
                    aborted_type is None or
                    not isinstance(aborted, aborted_type)
                ):
                    if context.task is None:
                        logger.warning(
                            f"Unhandled (in 'resemble.integrations.mailgun.v1.Message.SendTask') {aborted}; propagating as 'Unknown'\n" +
                            ''.join(traceback.format_exception(aborted))
                        )
                    raise resemble.aio.aborted.SystemAborted(
                        resemble.v1alpha1.errors_pb2.Unknown(),
                        # TODO(benh): consider whether or not we want to
                        # include the 'package.service.method' which may
                        # get concatenated together forming a kind of
                        # "stack trace"; while it's super helpful for
                        # debugging, it does expose implementation
                        # information.
                        message=f"unhandled (in 'resemble.integrations.mailgun.v1.Message.SendTask') {aborted}"
                    )

            raise
        except asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except BaseException as exception:
            if context.task is None:
                logger.warning(
                    f"Unhandled (in 'resemble.integrations.mailgun.v1.Message.SendTask') {type(exception).__name__}: {exception}; propagating as 'Unknown'\n" +
                    ''.join(traceback.format_exception(exception))
                )
            raise resemble.aio.aborted.SystemAborted(
                resemble.v1alpha1.errors_pb2.Unknown(),
                # TODO(benh): consider whether or not we want to
                # include the 'package.service.method' which may
                # get concatenated together forming a kind of
                # "stack trace"; while it's super helpful for
                # debugging, it does expose implementation
                # information.
                message=f"unhandled (in 'resemble.integrations.mailgun.v1.Message.SendTask') {type(exception).__name__}: {exception}"
            )

    async def _SendTask(
        self,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest,
        context: WriterContext,
        *,
        validating_effects: bool,
        grpc_context: Optional[grpc.aio.ServicerContext] = None,
    ) -> resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse:
        # Try to verify the token if a token verifier exists.
        context.auth = await self._maybe_verify_token(context._headers)

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        idempotent_mutation = self._state_manager.check_for_idempotent_mutation(
            context
        )

        if idempotent_mutation is not None:
            response = resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse()
            response.ParseFromString(idempotent_mutation.response)
            return response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__service_name__, context.actor_id
                )
            # TODO: this loads a `state` object even when we know we're calling a constructor,
            # in which case (by definition) there is no state. Can we (and is it worth
            # the effort to) change this call to avoid that overhead in that case?
            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                transaction=transaction,
                from_constructor=False,
                requires_constructor=True
            ) as (state, writer):

                await self._maybe_authorize(
                    method_name=
                    'resemble.integrations.mailgun.v1.Message.SendTask',
                    headers=context._headers,
                    auth=context.auth,
                    state=state,
                    request=request,
                )

                effects = await self.__SendTask(
                    context,
                    state,
                    request,
                    validating_effects=validating_effects,
                )

                if effects.tasks is not None:
                    # NOTE: we validate tasks added as part of a
                    # transaction when we prepare.
                    if transaction is None:
                        await self.tasks_dispatcher.validate(
                            effects.tasks
                        )

                await writer.complete(effects)

                if effects.tasks is not None:
                    if transaction is None:
                        self.tasks_dispatcher.dispatch(effects.tasks)
                    else:
                        assert all(
                            task.task_id.service == transaction.service
                            for task in effects.tasks
                        ), 'Task service does not match transaction service'
                        transaction.tasks.extend(effects.tasks)

                # TODO: We need a single `Effects` superclass for all methods, so we
                # would need to make it "partially" generic (with per-method subclasses
                # filling out the rest of the generic parameters) in order to fix this.
                return effects.response  # type: ignore[return-value]

    async def _schedule_SendTask(
        self,
        *,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest,
        headers: Headers,
        grpc_context: grpc.aio.ServicerContext,
    ) -> tuple[WriterContext, resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse]:
        context: WriterContext = self.create_context(
            headers=headers,
            context_type=WriterContext,
        )
        response = resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse()

        # Check if we already have performed this mutation!
        #
        # We do this _before_ calling 'transactionally()' because
        # if this call is for a transaction method _and_ we've
        # already performed the transaction then we don't want to
        # become a transaction participant (again) we just want to
        # return the transaction's response.
        idempotent_mutation = self._state_manager.check_for_idempotent_mutation(
            context
        )

        if idempotent_mutation is not None:
            response.ParseFromString(idempotent_mutation.response)

            # We should have only scheduled a single task!
            assert len(idempotent_mutation.task_ids) == 1
            assert grpc_context is not None
            grpc_context.set_trailing_metadata(
                grpc_context.trailing_metadata() +
                (
                    (
                        resemble.aio.headers.TASK_ID_UUID,
                        str(uuid.UUID(bytes=idempotent_mutation.task_ids[0].task_uuid))
                    ),
                )
            )

            return context, response

        async with self._state_manager.transactionally(
            context, self.tasks_dispatcher
        ) as transaction:
            if transaction is not None:
                context.participants.add(
                    self._servicer.__service_name__, context.actor_id
                )
            async with self._state_manager.writer(
                context,
                self._servicer.__state_type__,
                transaction=transaction,
                from_constructor=False,
                requires_constructor=True
            ) as (state, writer):

                # Try to verify the token if a token verifier exists.
                context.auth = await self._maybe_verify_token(headers)

                # Check if we are allowed schedule the task.
                await self._maybe_authorize(
                    method_name=
                    'resemble.integrations.mailgun.v1.Message.SendTask',
                    headers=context._headers,
                    auth=context.auth,
                    state=state,
                    request=request,
                )

                task = MessageServicerTasksStub(
                    context=context,
                    actor_id=context.actor_id,
                ).SendTask(
                    request,
                    schedule=context._headers.task_schedule,
                )

                effects = Effects(
                    response=response,
                    state=state,
                    tasks=[task],
                )

                assert effects.tasks is not None

                # NOTE: we validate tasks added as part of a
                # transaction when we prepare.
                if transaction is None:
                    await self.tasks_dispatcher.validate(effects.tasks)

                await writer.complete(effects)

                if transaction is None:
                    self.tasks_dispatcher.dispatch(effects.tasks)
                else:
                    assert task.task_id.service == transaction.service
                    transaction.tasks.extend(effects.tasks)

                assert grpc_context is not None

                grpc_context.set_trailing_metadata(
                    grpc_context.trailing_metadata() +
                    (
                        (
                            resemble.aio.headers.TASK_ID_UUID,
                            str(uuid.UUID(bytes=task.task_id.task_uuid))
                        ),
                    )
                )

                return context, response

        return context, response


    @maybe_run_method_twice_to_validate_effects
    async def SendTask(
        self,
        validating_effects: bool,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest,
        grpc_context: grpc.aio.ServicerContext,
    ) -> resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse:
        context: Optional[Context] = None
        try:
            headers = Headers.from_grpc_context(grpc_context)
            if headers.task_schedule is not None:
                context, response = await self._schedule_SendTask(
                    headers=headers,
                    request=request,
                    grpc_context=grpc_context,
                )
                return response

            context = self.create_context(
                headers=headers,
                context_type=WriterContext,
            )
            assert context is not None

            return await self._SendTask(
                request,
                context,
                validating_effects=validating_effects,
                grpc_context=grpc_context,
            )
        except EffectValidationRetry:
            # Doing effect validation, just let this propagate.
            raise
        except resemble.aio.aborted.Aborted as aborted:
            await grpc_context.abort_with_status(
                rpc_status.to_status(aborted.to_status())
            )
            raise  # Unreachable but necessary for mypy.
        except asyncio.CancelledError:
            # It's pretty normal for an RPC to be cancelled; it's not useful to
            # print a stack trace.
            raise
        except:
            # Print the exception stack trace for easier debugging. Note
            # that we don't include the stack trace in an error message
            # for the same reason that gRPC doesn't do so by default,
            # see https://github.com/grpc/grpc/issues/14897, but since this
            # should only get logged on the server side it is safe.
            logger.warning(
                'Unhandled exception\n' +
                ''.join(traceback.format_exc())
            )

            # Re-raise the exception for gRPC to handle!
            raise
        finally:
            if context is not None and context.transaction_id is not None:
                # Propagate transaction participants.
                #
                # NOTE: we'll want to propagate participants if we're
                # in a transaction method once we support nested
                # transactions, but for now we only need to propagate
                # participants when we're not in a transaction method.
                if not isinstance(context, TransactionContext):
                    grpc_context.set_trailing_metadata(
                        grpc_context.trailing_metadata() +
                        context.participants.to_grpc_metadata()
                    )

    async def _maybe_authorize(
        self,
        *,
        method_name: str,
        headers: Headers,
        auth: Optional[resemble.aio.auth.Auth],
        state: Optional[MessageStateType] = None,
        request: Optional[MessageRequestTypes] = None,
    ) -> None:
        """Check authorization decision for the given method, using available
        information.

        Raises `PermissionDenied` in case authorizer is present but the request
        is not authorized.
        """
        # To authorize internal calls, we use an internal magic token.
        if headers.bearer_token == __internal_magic_token__:
            return

        if self._authorizer is None:
            return

        if not self._authorizer.can_authorize(method_name):
            # Better safe than sorry.
            raise resemble.aio.aborted.SystemAborted(
                resemble.v1alpha1.errors_pb2.PermissionDenied(),
                message=
                    f"Authorizer does not support method '{method_name}'"
            )

        # If there is state, ensure that we don't accidentally mutate it.
        if state is not None:
            # TODO: If we can read-only a state object we can avoid a copy here.
            state_copy = MessageStateType()
            state_copy.CopyFrom(state)
            state = state_copy

        # Create context for the authorizer. This is a `ReaderContext`
        # independently of the calling context.
        with self.use_context(
            headers=(
                # Get headers suitable for doing authorization.
                headers.copy_for_token_verification_and_authorization()
            ),
            context_type=ReaderContext,
        ) as context:
            context.auth = auth

            # Get the authorizer decision.
            authorization_decision = await self._authorizer.authorize(
                method_name=method_name,
                context=context,
                state=state,
                request=request,
            )

        # Enforce correct authorizer decision type.
        try:
            assert_type(
                authorization_decision,
                [
                    resemble.v1alpha1.errors_pb2.Ok,
                    resemble.v1alpha1.errors_pb2.Unauthenticated,
                    resemble.v1alpha1.errors_pb2.PermissionDenied,
                ]
            )
        except TypeError as e:
            # Recast the exception to provide more context.
            authorizer_type = f"{type(self._authorizer).__module__}.{type(self._authorizer).__name__}"
            raise TypeError(
                f"Authorizer '{authorizer_type}' "
                f"returned unexpected type '{type(authorization_decision).__name__}' "
                f"for method '{method_name}' on '{headers.actor_id}'"
            ) from e

        # If the decision is not `True`, raise a `SystemAborted` with either a
        # `PermissionDenied` error (in case of `False`) or an `Unauthenticated`
        # error.
        if not isinstance(authorization_decision, resemble.v1alpha1.errors_pb2.Ok):
            if isinstance(authorization_decision, resemble.v1alpha1.errors_pb2.Unauthenticated):
                logger.warning(
                    "Unauthenticated call to "
                    f"'{method_name}' on '{headers.actor_id}'"
                )

            raise resemble.aio.aborted.SystemAborted(
                authorization_decision,
                message=
                f'You are not authorized to call {method_name} on '
                f"'{headers.actor_id}'"
            )

    async def _maybe_verify_token(self, headers) -> Optional[resemble.aio.auth.Auth]:
        """Verify the bearer token and if a token verifier is present.

        Returns the (optional) `resemble.aio.auth.Auth` object
        produced by the token verifier if the token can be verified.
        """
        auth: Optional[resemble.aio.auth.Auth] = None

        if self._token_verifier is not None and headers.bearer_token is not None:
            if headers.bearer_token == __internal_magic_token__:
                return auth

            with self.use_context(
                headers=(
                    # Get headers suitable for doing token verification.
                    headers.copy_for_token_verification_and_authorization()
                ),
                context_type=ReaderContext,
            ) as context:
                auth = await self._token_verifier.verify_token(
                    context=context,
                    token=headers.bearer_token,
                )

        return auth


class _MessageStub(Stub):

    def __init__(
        self,
        *,
        context_or_workflow: Context | Workflow,
        actor_id: ActorId,
        bearer_token: Optional[str] = None,
    ):
        context: Optional[Context] = None

        if isinstance(context_or_workflow, Context):
            context = context_or_workflow

            # ISSUE(https://github.com/reboot-dev/respect/issues/2258)
            # For now we do not explicitly pass on the bearer token here, as it
            # might have security implications to just pass on the token to any
            # service we are calling.

        elif isinstance(context_or_workflow, Workflow):
            bearer_token = context_or_workflow.bearer_token

        super().__init__(
            channel_manager=context_or_workflow.channel_manager,
            idempotency_manager=context_or_workflow,
            service_name='resemble.integrations.mailgun.v1.Message',
            actor_id=actor_id,
            context=context,
            bearer_token=bearer_token,
        )

        channel = self._channel_manager.get_channel_for(
            'resemble.integrations.mailgun.v1.Message', actor_id
        )
        self._stub = resemble.integrations.mailgun.v1.mailgun_pb2_grpc.MessageStub(channel)


class MessageReaderStub(_MessageStub):

    def __init__(
        self,
        context_or_workflow: ReaderContext | WriterContext | TransactionContext | WorkflowContext | Workflow,
        *,
        actor_id: ActorId,
        bearer_token: Optional[str] = None,
    ):
        assert_type(
            context_or_workflow,
            [ReaderContext, WriterContext, TransactionContext, WorkflowContext, Workflow]
        )
        super().__init__(
            context_or_workflow=context_or_workflow,
            actor_id=actor_id,
            bearer_token=bearer_token,
        )

    # Message specific methods:




class MessageWriterStub(_MessageStub):

    def __init__(
        self,
        context_or_workflow: TransactionContext | WorkflowContext | Workflow,
        *,
        actor_id: ActorId,
        bearer_token: Optional[str] = None,
    ):
        assert_type(context_or_workflow, [TransactionContext, WorkflowContext, Workflow])
        super().__init__(
            context_or_workflow=context_or_workflow,
            actor_id=actor_id,
            bearer_token=bearer_token,
        )

    # Message specific methods:
    async def Send(
        self,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
        bearer_token: Optional[str] = None,
    ) -> resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse:
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='Send',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'Send',
                self._stub.Send,
                request,
                unary=True,
                reader=False,
                response_type=resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse,
                aborted_type=Message.SendAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call

    async def SendTask(
        self,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
        bearer_token: Optional[str] = None,
    ) -> resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse:
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='SendTask',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'SendTask',
                self._stub.SendTask,
                request,
                unary=True,
                reader=False,
                response_type=resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse,
                aborted_type=Message.SendTaskAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call


class MessageWorkflowStub(_MessageStub):

    def __init__(
        self,
        *,
        context_or_workflow: WorkflowContext | Workflow,
        actor_id: ActorId,
        bearer_token: Optional[str] = None,
    ):
        assert_type(context_or_workflow, [WorkflowContext, Workflow])
        super().__init__(
            context_or_workflow=context_or_workflow,
            actor_id=actor_id,
            bearer_token=bearer_token,
        )

    # Message specific methods:
    async def Send(
        self,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
        bearer_token: Optional[str] = None,
    ) -> resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='Send',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'Send',
                self._stub.Send,
                request,
                unary=True,
                reader=False,
                response_type=resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse,
                aborted_type=Message.SendAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call

    async def SendTask(
        self,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
        bearer_token: Optional[str] = None,
    ) -> resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='SendTask',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'SendTask',
                self._stub.SendTask,
                request,
                unary=True,
                reader=False,
                response_type=resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse,
                aborted_type=Message.SendTaskAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, Awaitable)
                return await call



class MessageTasksStub(_MessageStub):

    def __init__(
        self,
        context_or_workflow: TransactionContext | WorkflowContext | Workflow,
        *,
        actor_id: ActorId,
        bearer_token: Optional[str] = None,
    ):
        assert_type(context_or_workflow, [TransactionContext, WorkflowContext, Workflow])
        super().__init__(
            context_or_workflow=context_or_workflow,
            actor_id=actor_id,
            bearer_token=bearer_token,
        )

    # Message specific methods:
    async def Send(
        self,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
        bearer_token: Optional[str] = None,
    ) -> tasks_pb2.TaskId:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='Send',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'Send',
                self._stub.Send,
                request,
                unary=True,
                reader=False,
                response_type=resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse,
                aborted_type=Message.SendAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, Awaitable)
                await call
                for (key, value) in await call.trailing_metadata():  # type: ignore[attr-defined]
                    if key == resemble.aio.headers.TASK_ID_UUID:
                        return tasks_pb2.TaskId(
                            service=self._headers.service_name,
                            actor_id=self._headers.actor_id,
                            task_uuid=uuid.UUID(value).bytes,
                        )
                raise resemble.aio.aborted.SystemAborted(
                    resemble.v1alpha1.errors_pb2.Internal(),
                    message='Trailing metadata missing for task schedule',
                )

    async def SendTask(
        self,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest,
        idempotency: Optional[Idempotency] = None,
        *,
        metadata: Optional[GrpcMetadata] = None,
        bearer_token: Optional[str] = None,
    ) -> tasks_pb2.TaskId:
        idempotency_key: Optional[str]
        with self._idempotency_manager.idempotently(
            service=self._headers.service_name,
            actor_id=self._headers.actor_id,
            method='SendTask',
            request=request,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:
            async with self._call(
                'SendTask',
                self._stub.SendTask,
                request,
                unary=True,
                reader=False,
                response_type=resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse,
                aborted_type=Message.SendTaskAborted,
                metadata=metadata,
                idempotency_key=idempotency_key,
                bearer_token=bearer_token,
            ) as call:
                assert isinstance(call, Awaitable)
                await call
                for (key, value) in await call.trailing_metadata():  # type: ignore[attr-defined]
                    if key == resemble.aio.headers.TASK_ID_UUID:
                        return tasks_pb2.TaskId(
                            service=self._headers.service_name,
                            actor_id=self._headers.actor_id,
                            task_uuid=uuid.UUID(value).bytes,
                        )
                raise resemble.aio.aborted.SystemAborted(
                    resemble.v1alpha1.errors_pb2.Internal(),
                    message='Trailing metadata missing for task schedule',
                )



class MessageServicerTasksStub(_MessageStub):

    def __init__(
        self,
        context: WriterContext,
        *,
        actor_id: ActorId,
        bearer_token: Optional[str] = None,
    ):
        assert_type(context, [WriterContext])
        super().__init__(
            context_or_workflow=context,
            actor_id=actor_id,
            bearer_token=bearer_token,
        )

    # Message specific methods:
    def Send(
        self,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest,
        *,
        schedule: Optional[datetime | timedelta] = None,
    ) -> TaskEffect:
        return TaskEffect(
            service_name='resemble.integrations.mailgun.v1.Message',
            actor_id=self._headers.actor_id,
            method_name='Send',
            request=request,
            schedule=(datetime.now() + schedule) if isinstance(
                schedule, timedelta
            ) else schedule,
        )

    def SendTask(
        self,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest,
        *,
        schedule: Optional[datetime | timedelta] = None,
    ) -> TaskEffect:
        return TaskEffect(
            service_name='resemble.integrations.mailgun.v1.Message',
            actor_id=self._headers.actor_id,
            method_name='SendTask',
            request=request,
            schedule=(datetime.now() + schedule) if isinstance(
                schedule, timedelta
            ) else schedule,
        )



class MessageAuthorizerMixin:
    """Mixin that provides a default implementation of the `can_authorize`
    method for all Message authorizers.
    """

    def can_authorize(self, method_name: str) -> bool:
        service_method_names = [
            'resemble.integrations.mailgun.v1.Message.Send',
            'resemble.integrations.mailgun.v1.Message.SendTask',
        ]
        return method_name in service_method_names


MessageStateType: TypeAlias = resemble.integrations.mailgun.v1.mailgun_pb2.MessageState
MessageRequestTypes: TypeAlias = \
        resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest \
        | resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest


class MessageAuthorizer(
    MessageAuthorizerMixin,
    resemble.aio.auth.authorizers.Authorizer[MessageStateType, MessageRequestTypes],
):
    StateType: TypeAlias = MessageStateType
    RequestTypes: TypeAlias = MessageRequestTypes

    async def authorize(
        self,
        *,
        method_name: str,
        context: ReaderContext,
        state: Optional[MessageStateType] = None,
        request: Optional[MessageRequestTypes] = None,
    ) -> resemble.aio.auth.authorizers.Authorizer.Decision:
        if method_name == 'resemble.integrations.mailgun.v1.Message.Send':
            return await self.Send(
                context,
                cast(resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest, request),
            )
        elif method_name == 'resemble.integrations.mailgun.v1.Message.SendTask':
            return await self.SendTask(
                context,
                cast(resemble.integrations.mailgun.v1.mailgun_pb2.MessageState, state),
                cast(resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest, request),
            )
        else:
            return resemble.v1alpha1.errors_pb2.PermissionDenied()

    async def Send(
        self,
        context: ReaderContext,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest,
    ) -> resemble.aio.auth.authorizers.Authorizer.Decision:
        return resemble.v1alpha1.errors_pb2.PermissionDenied()

    async def SendTask(
        self,
        context: ReaderContext,
        state: resemble.integrations.mailgun.v1.mailgun_pb2.MessageState,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest,
    ) -> resemble.aio.auth.authorizers.Authorizer.Decision:
        return resemble.v1alpha1.errors_pb2.PermissionDenied()


    # Off-the shelf authorizers for Message:

    class AllowAllIfAuthenticated(
        MessageAuthorizerMixin,
        resemble.aio.auth.authorizers.Authorizer,
    ):
        """An authorizer for Message that allows all requests if the
        caller is authenticated.
        """
        async def authorize(
            self,
            *,
            method_name: str,
            context: ReaderContext,
            state: Optional[MessageStateType] = None,
            request: Optional[MessageRequestTypes] = None,
        ) -> resemble.aio.auth.authorizers.Authorizer.Decision:
            if context.auth is None:
                return resemble.v1alpha1.errors_pb2.Unauthenticated()

            return resemble.v1alpha1.errors_pb2.Ok()


class MessageServicer(Servicer):
    Authorizer: TypeAlias = MessageAuthorizer

    __service_name__ = 'resemble.integrations.mailgun.v1.Message'
    __file_descriptor__ = resemble.integrations.mailgun.v1.mailgun_pb2.DESCRIPTOR
    __state_type__ = resemble.integrations.mailgun.v1.mailgun_pb2.MessageState

    def __init__(self):
        # NOTE: need to hold on to the middleware so we can do inline
        # writes (see 'self.write(...)').
        #
        # Because '_middleware' is not really private this does mean
        # users may do possibly dangerous things, but this is no more
        # likely given they could have already overridden
        # 'create_middleware()'.
        self._middleware: Optional[MessageServicerMiddleware] = None

    def create_middleware(
        self,
        *,
        application_id: ApplicationId,
        state_manager: StateManager,
        channel_manager: _ChannelManager,
        tasks_cache: TasksCache,
        token_verifier: Optional[resemble.aio.auth.token_verifiers.TokenVerifier],
        effect_validation: EffectValidation,
        ready: asyncio.Event,
    ) -> MessageServicerMiddleware:
        self._middleware = MessageServicerMiddleware(
            servicer=self,
            application_id=application_id,
            state_manager=state_manager,
            channel_manager=channel_manager,
            tasks_cache=tasks_cache,
            token_verifier=token_verifier,
            effect_validation=effect_validation,
            ready=ready,
        )
        return self._middleware

    def authorizer(self) -> Optional[resemble.aio.auth.authorizers.Authorizer]:
        return None

    def token_verifier(self) -> Optional[resemble.aio.auth.token_verifiers.TokenVerifier]:
        return None

    def lookup(
        self,
        *,
        bearer_token: Optional[str] = None,
    ) -> Message.WeakReference:
        # TODO(benh): it's still possible that the `bearer_token` from
        # the context will not be sufficient for making a
        # call/schedule due to a user defined authorizer; consider
        # creating a "per middleware" bearer token that always works
        # for calls from `self.lookup()`.
        context = Context.get()

        if context is None:
            raise RuntimeError(
                'Missing asyncio context variable `context`; '
                'are you using this class without Resemble?'
            )

        return Message.WeakReference(
            actor_id=context.actor_id,
            bearer_token=bearer_token or context.bearer_token,
        )

    def schedule(
        self,
        *,
        when: Optional[datetime | timedelta] = None,
    ):
        class Tasks:
            def __init__(self, when: Optional[datetime | timedelta] = None):
                self._when = when

            # Message specific methods:
            def SendTask(
                self,
                __context__: WriterContext,
                *,
                send_request: Optional[resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest] = None,
            ) -> TaskEffect:
                if send_request is not None and not isinstance(
                    send_request,
                    resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest,
                ):
                    raise TypeError(
                        f"Can not construct protobuf message of type "
                        f"'resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest': field 'send_request' is not "
                        f"of required type 'resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest'"
                    )
                # TODO: mypy-protobuf declares that
                # `google.protobuf.message.Message` constructor arguments
                # are always non-None, when in reality they are optional.
                request = resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest(
                    send_request=send_request,  # type: ignore[arg-type]
                )
                return MessageServicerTasksStub(
                    context=__context__,
                    actor_id=__context__.actor_id,
                ).SendTask(
                    request,
                    schedule=self._when,
                )


        return Tasks(when)

    class Effects(resemble.aio.state_managers.Effects):
        def __init__(
            self,
            *,
            state: resemble.integrations.mailgun.v1.mailgun_pb2.MessageState,
            response: Optional[google.protobuf.message.Message] = None,
            tasks: Optional[list[TaskEffect]] = None,
            _interleaved_upserts: Optional[list[tuple[str, Optional[bytes]]]] = None,
        ):
            assert_type(state, [resemble.integrations.mailgun.v1.mailgun_pb2.MessageState])

            super().__init__(state=state, response=response, tasks=tasks, _interleaved_upserts=_interleaved_upserts)

    class SendEffects(Effects):
        def __init__(
            self,
            *,
            state: resemble.integrations.mailgun.v1.mailgun_pb2.MessageState,
            response: resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse,
            tasks: Optional[list[TaskEffect]] = None,
            _interleaved_upserts: Optional[list[tuple[str, Optional[bytes]]]] = None,
        ):
            assert_type(state, [resemble.integrations.mailgun.v1.mailgun_pb2.MessageState])
            assert_type(response, [resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse])

            super().__init__(state=state, response=response, tasks=tasks, _interleaved_upserts=_interleaved_upserts)


    class SendTaskEffects(Effects):
        def __init__(
            self,
            *,
            state: resemble.integrations.mailgun.v1.mailgun_pb2.MessageState,
            response: resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse,
            tasks: Optional[list[TaskEffect]] = None,
            _interleaved_upserts: Optional[list[tuple[str, Optional[bytes]]]] = None,
        ):
            assert_type(state, [resemble.integrations.mailgun.v1.mailgun_pb2.MessageState])
            assert_type(response, [resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse])

            super().__init__(state=state, response=response, tasks=tasks, _interleaved_upserts=_interleaved_upserts)



    class InlineWriterCallable(typing.Protocol):
        async def __call__(
            self,
            context: WriterContext,
            state: resemble.integrations.mailgun.v1.mailgun_pb2.MessageState
        ) -> MessageServicer.Effects:
            ...

    async def read(
        self, context: TransactionContext | WorkflowContext
    ) -> resemble.integrations.mailgun.v1.mailgun_pb2.MessageState:
        """Read the current state within a transaction."""
        assert_type(context, [TransactionContext, WorkflowContext])

        if self._middleware is None:
            raise RuntimeError(
                'Resemble middleware was not created; '
                'are you using this class without Resemble?'
            )

        return await self._middleware._state_manager.read(
            context, self.__state_type__
        )

    async def write(
        self,
        context: TransactionContext | WorkflowContext,
        writer: InlineWriterCallable,
        __options__: Optional[resemble.aio.call.Options] = None,
    ):
        """Perform an "inline write" within a transaction or task."""
        assert_type(context, [TransactionContext, WorkflowContext])

        if self._middleware is None:
            raise RuntimeError(
                'Resemble middleware was not created; '
                'are you using this class without Resemble?'
            )

        idempotency: Optional[resemble.aio.idempotency.Idempotency] = None
        metadata: Optional[resemble.aio.types.GrpcMetadata] = None

        if __options__ is not None:
            if __options__.idempotency_key is not None:
                idempotency = Idempotency(key=__options__.idempotency_key)
            elif __options__.idempotency_alias is not None:
                idempotency = Idempotency(alias=__options__.idempotency_alias)
            if __options__.metadata is not None:
                metadata = __options__.metadata

        if metadata is None:
            metadata = ()

        headers = Headers(
            application_id=context.application_id,
            service_name=self.__service_name__,
            actor_id=context.actor_id,
            transaction_id=context.transaction_id,
            transaction_coordinator_service=context
            .transaction_coordinator_service,
            transaction_coordinator_actor_id=context
            .transaction_coordinator_actor_id
        )

        metadata += headers.to_grpc_metadata()

        idempotency_key: Optional[str]
        with context.idempotently(
            service=self.__service_name__,
            actor_id=context.actor_id,
            method=None,
            request=None,
            metadata=metadata,
            idempotency=idempotency,
        ) as idempotency_key:

            if any(t[0] == resemble.aio.headers.IDEMPOTENCY_KEY_HEADER for t in metadata):
                raise ValueError(
                    f"Do not set '{resemble.aio.headers.IDEMPOTENCY_KEY_HEADER}' metadata yourself"
                )

            if idempotency_key is not None:
                metadata += (
                    (resemble.aio.headers.IDEMPOTENCY_KEY_HEADER, idempotency_key),
                )

            with self._middleware.use_context(
                headers=Headers.from_grpc_metadata(metadata),
                context_type=WriterContext,
            ) as writer_context:
                # Check if we already have performed this mutation!
                #
                # We do this _before_ calling 'transactionally()' because
                # if this call is for a transaction method _and_ we've
                # already performed the transaction then we don't want to
                # become a transaction participant (again) we just want to
                # return the transaction's response.
                idempotent_mutation = (
                    self._middleware._state_manager.check_for_idempotent_mutation(
                        writer_context
                    )
                )

                if idempotent_mutation is not None:
                    # NOTE: invariant for now is that inline writers can't
                    # have responses which we also check for below.
                    assert len(idempotent_mutation.response) == 0
                    return

                async with self._middleware._state_manager.transactionally(
                    writer_context, self._middleware.tasks_dispatcher
                ) as transaction:
                    async with self._middleware._state_manager.writer(
                        writer_context,
                        self.__state_type__,
                        transaction=transaction,
                    ) as (state, state_manager_writer):
                        effects: MessageServicer.Effects = await writer(
                            context=writer_context, state=state
                        )

                        assert_type(
                            effects,
                            [MessageServicer.Effects],
                            may_be_subclass=False,
                        )

                        # TODO(benh): figure out if we want to support
                        # returning a response because we don't know how
                        # to parse it from an idempotent mutation.
                        if effects.response is not None:
                            raise ValueError(
                                'Returning a response from an inline writer '
                                'is not supported'
                            )

                        # All idempotent mutations require a reponse, so
                        # for now we just use `Empty`.
                        effects.response = empty_pb2.Empty()

                        if effects.tasks is not None:
                            # NOTE: we validate tasks added as part of a
                            # transaction when we prepare.
                            if transaction is None:
                                await self._middleware.tasks_dispatcher.validate(
                                    effects.tasks
                                )

                        await state_manager_writer.complete(effects)

                        if effects.tasks is not None:
                            if transaction is None:
                                self._middleware.tasks_dispatcher.dispatch(effects.tasks)
                            else:
                                assert all(
                                    task.task_id.service == transaction.service
                                    for task in effects.tasks
                                ), 'Task service does not match transaction service'
                                transaction.tasks.extend(effects.tasks)

    # Message specific methods:
    @abstractmethod
    async def Send(
        self,
        context: WriterContext,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest,
    ) -> SendEffects:
        raise NotImplementedError

    @abstractmethod
    async def SendTask(
        self,
        context: WriterContext,
        state: resemble.integrations.mailgun.v1.mailgun_pb2.MessageState,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest,
    ) -> SendTaskEffects:
        raise NotImplementedError


class Message:

    Interface: TypeAlias = MessageServicer

    Effects: TypeAlias = MessageServicer.Effects

    Authorizer: TypeAlias = MessageAuthorizer

    class SendFuture:
        """Represents a future corresponding to a task running on the
        actor. Note that this is not a coroutine because we are trying
        to convey the semantics that the task is already running (or
        will soon be) and thus we are just giving you a (distributed)
        future to it.
        """

        def __init__(
            self,
            context_or_workflow: Context | Workflow,
            *,
            task_id: tasks_pb2.TaskId,
        ) -> None:
            self._channel_manager = context_or_workflow.channel_manager
            self._task_id = task_id

        @property
        def task_id(self) -> tasks_pb2.TaskId:
            return self._task_id

        def __await__(self) -> Generator[None, None, resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse]:
            """Awaits for task to finish and returns it's response."""

            async def wait_for_task() -> resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse:
                channel = self._channel_manager.get_channel_from_service_name(
                    self._task_id.service, self._task_id.actor_id
                )

                stub = tasks_pb2_grpc.TasksStub(channel)

                try:
                    wait_for_task_response = await stub.Wait(
                        tasks_pb2.WaitRequest(task_id=self._task_id),
                        metadata=Headers(
                            service_name=self._task_id.service,
                            actor_id=self._task_id.actor_id,
                            # TODO(benh): will eventually be necessary to
                            # disambiguate between applications.
                            application_id=None,
                        ).to_grpc_metadata(),
                    )
                except grpc.aio.AioRpcError as error:
                    if error.code() == grpc.StatusCode.NOT_FOUND:
                        raise resemble.aio.aborted.SystemAborted(
                            resemble.v1alpha1.errors_pb2.UnknownTask()
                        ) from None

                    raise resemble.aio.aborted.SystemAborted.from_grpc_aio_rpc_error(
                        error
                    ) from None
                else:
                    response = resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse()

                    if (
                        wait_for_task_response.response.TypeName() !=
                        response.DESCRIPTOR.full_name
                    ):
                        raise resemble.aio.aborted.SystemAborted(
                            resemble.v1alpha1.errors_pb2.InvalidArgument(),
                            message=
                            f"task with UUID {str(uuid.UUID(bytes=self._task_id.task_uuid))} "
                            f"has a response of type '{wait_for_task_response.response.TypeName()}' "
                            "but expecting type 'resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse'; "
                            "are you waiting on a task of the correct method?",
                        ) from None

                    wait_for_task_response.response.Unpack(response)
                    return response

            return wait_for_task().__await__()

    SendEffects: TypeAlias = Interface.SendEffects

    class SendAborted(resemble.aio.aborted.Aborted):


        Error = Union[
            resemble.aio.aborted.GrpcError,
            resemble.aio.aborted.ResembleError,
        ]

        METHOD_ERROR_TYPES: list[type[Error]] = [
        ]

        ERROR_TYPES: list[type[Error]] = (
            METHOD_ERROR_TYPES +
            resemble.aio.aborted.GRPC_ERROR_TYPES +
            resemble.aio.aborted.RESEMBLE_ERROR_TYPES
        )

        _error: Error
        _code: grpc.StatusCode
        _message: Optional[str]

        def __init__(
            self,
            error:  resemble.aio.aborted.GrpcError,
            *,
            message: Optional[str] = None,
            # Do not set this value when constructing in order to
            # raise. This is only used internally when constructing
            # from aborted calls.
            error_types: list[type[Error]] = (
                METHOD_ERROR_TYPES + resemble.aio.aborted.GRPC_ERROR_TYPES
            ),
        ):
            super().__init__()

            assert_type(error, error_types)

            self._error = error

            code = self.grpc_status_code_from_error(self._error)

            if code is None:
                # Must be a Resemble specific or declared method error.
                code = grpc.StatusCode.ABORTED

            self._code = code

            self._message = message

        @property
        def error(self) -> Error:
            return self._error

        @property
        def code(self) -> grpc.StatusCode:
            return self._code

        @property
        def message(self) -> Optional[str]:
            return self._message

        @classmethod
        def from_status(cls, status: status_pb2.Status):
            error = cls.error_from_google_rpc_status_details(
                status,
                cls.ERROR_TYPES,
            )

            message = status.message if len(status.message) > 0 else None

            if error is not None:
                return cls(error, message=message, error_types=cls.ERROR_TYPES)

            error = cls.error_from_google_rpc_status_code(status)

            assert error is not None

            # TODO(benh): also consider getting the type names from
            # `status.details` and including that in `message` to make
            # debugging easier.

            return cls(error, message=message)

        @classmethod
        def from_grpc_aio_rpc_error(cls, aio_rpc_error: grpc.aio.AioRpcError):
            return cls(
                cls.error_from_grpc_aio_rpc_error(aio_rpc_error),
                message=aio_rpc_error.details(),
            )

        @classmethod
        def is_declared_error(cls, message: google.protobuf.message.Message) -> bool:
            return False

    class SendTaskFuture:
        """Represents a future corresponding to a task running on the
        actor. Note that this is not a coroutine because we are trying
        to convey the semantics that the task is already running (or
        will soon be) and thus we are just giving you a (distributed)
        future to it.
        """

        def __init__(
            self,
            context_or_workflow: Context | Workflow,
            *,
            task_id: tasks_pb2.TaskId,
        ) -> None:
            self._channel_manager = context_or_workflow.channel_manager
            self._task_id = task_id

        @property
        def task_id(self) -> tasks_pb2.TaskId:
            return self._task_id

        def __await__(self) -> Generator[None, None, resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse]:
            """Awaits for task to finish and returns it's response."""

            async def wait_for_task() -> resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse:
                channel = self._channel_manager.get_channel_from_service_name(
                    self._task_id.service, self._task_id.actor_id
                )

                stub = tasks_pb2_grpc.TasksStub(channel)

                try:
                    wait_for_task_response = await stub.Wait(
                        tasks_pb2.WaitRequest(task_id=self._task_id),
                        metadata=Headers(
                            service_name=self._task_id.service,
                            actor_id=self._task_id.actor_id,
                            # TODO(benh): will eventually be necessary to
                            # disambiguate between applications.
                            application_id=None,
                        ).to_grpc_metadata(),
                    )
                except grpc.aio.AioRpcError as error:
                    if error.code() == grpc.StatusCode.NOT_FOUND:
                        raise resemble.aio.aborted.SystemAborted(
                            resemble.v1alpha1.errors_pb2.UnknownTask()
                        ) from None

                    raise resemble.aio.aborted.SystemAborted.from_grpc_aio_rpc_error(
                        error
                    ) from None
                else:
                    response = resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse()

                    if (
                        wait_for_task_response.response.TypeName() !=
                        response.DESCRIPTOR.full_name
                    ):
                        raise resemble.aio.aborted.SystemAborted(
                            resemble.v1alpha1.errors_pb2.InvalidArgument(),
                            message=
                            f"task with UUID {str(uuid.UUID(bytes=self._task_id.task_uuid))} "
                            f"has a response of type '{wait_for_task_response.response.TypeName()}' "
                            "but expecting type 'resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse'; "
                            "are you waiting on a task of the correct method?",
                        ) from None

                    wait_for_task_response.response.Unpack(response)
                    return response

            return wait_for_task().__await__()

    SendTaskEffects: TypeAlias = Interface.SendTaskEffects

    class SendTaskAborted(resemble.aio.aborted.Aborted):


        Error = Union[
            resemble.aio.aborted.GrpcError,
            resemble.aio.aborted.ResembleError,
        ]

        METHOD_ERROR_TYPES: list[type[Error]] = [
        ]

        ERROR_TYPES: list[type[Error]] = (
            METHOD_ERROR_TYPES +
            resemble.aio.aborted.GRPC_ERROR_TYPES +
            resemble.aio.aborted.RESEMBLE_ERROR_TYPES
        )

        _error: Error
        _code: grpc.StatusCode
        _message: Optional[str]

        def __init__(
            self,
            error:  resemble.aio.aborted.GrpcError,
            *,
            message: Optional[str] = None,
            # Do not set this value when constructing in order to
            # raise. This is only used internally when constructing
            # from aborted calls.
            error_types: list[type[Error]] = (
                METHOD_ERROR_TYPES + resemble.aio.aborted.GRPC_ERROR_TYPES
            ),
        ):
            super().__init__()

            assert_type(error, error_types)

            self._error = error

            code = self.grpc_status_code_from_error(self._error)

            if code is None:
                # Must be a Resemble specific or declared method error.
                code = grpc.StatusCode.ABORTED

            self._code = code

            self._message = message

        @property
        def error(self) -> Error:
            return self._error

        @property
        def code(self) -> grpc.StatusCode:
            return self._code

        @property
        def message(self) -> Optional[str]:
            return self._message

        @classmethod
        def from_status(cls, status: status_pb2.Status):
            error = cls.error_from_google_rpc_status_details(
                status,
                cls.ERROR_TYPES,
            )

            message = status.message if len(status.message) > 0 else None

            if error is not None:
                return cls(error, message=message, error_types=cls.ERROR_TYPES)

            error = cls.error_from_google_rpc_status_code(status)

            assert error is not None

            # TODO(benh): also consider getting the type names from
            # `status.details` and including that in `message` to make
            # debugging easier.

            return cls(error, message=message)

        @classmethod
        def from_grpc_aio_rpc_error(cls, aio_rpc_error: grpc.aio.AioRpcError):
            return cls(
                cls.error_from_grpc_aio_rpc_error(aio_rpc_error),
                message=aio_rpc_error.details(),
            )

        @classmethod
        def is_declared_error(cls, message: google.protobuf.message.Message) -> bool:
            return False


    @classmethod
    async def Send(
        cls,
        __actor_id__: ActorId,
        __context_or_workflow__: TransactionContext | WorkflowContext | Workflow,
        __options__: Optional[resemble.aio.call.Options] = None,
        *,
        recipient: Optional[str] = None,
        sender: Optional[str] = None,
        subject: Optional[str] = None,
        domain: Optional[str] = None,
        text: Optional[str] = None,
        html: Optional[str] = None,
    ) -> tuple[Message.WeakReference, resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse]:
        if recipient is not None and not isinstance(
            recipient,
            str,
        ):
            raise TypeError(
                f"Can not construct protobuf message of type "
                f"'resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest': field 'recipient' is not "
                f"of required type 'str'"
            )
        if sender is not None and not isinstance(
            sender,
            str,
        ):
            raise TypeError(
                f"Can not construct protobuf message of type "
                f"'resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest': field 'sender' is not "
                f"of required type 'str'"
            )
        if subject is not None and not isinstance(
            subject,
            str,
        ):
            raise TypeError(
                f"Can not construct protobuf message of type "
                f"'resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest': field 'subject' is not "
                f"of required type 'str'"
            )
        if domain is not None and not isinstance(
            domain,
            str,
        ):
            raise TypeError(
                f"Can not construct protobuf message of type "
                f"'resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest': field 'domain' is not "
                f"of required type 'str'"
            )
        if text is not None and not isinstance(
            text,
            str,
        ):
            raise TypeError(
                f"Can not construct protobuf message of type "
                f"'resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest': field 'text' is not "
                f"of required type 'str'"
            )
        if html is not None and not isinstance(
            html,
            str,
        ):
            raise TypeError(
                f"Can not construct protobuf message of type "
                f"'resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest': field 'html' is not "
                f"of required type 'str'"
            )
        # TODO: mypy-protobuf declares that
        # `google.protobuf.message.Message` constructor arguments are
        # always non-None, when in reality they are optional.
        request = resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest(
            recipient=recipient,  # type: ignore[arg-type]
            sender=sender,  # type: ignore[arg-type]
            subject=subject,  # type: ignore[arg-type]
            domain=domain,  # type: ignore[arg-type]
            text=text,  # type: ignore[arg-type]
            html=html,  # type: ignore[arg-type]
        )
        idempotency: Optional[resemble.aio.idempotency.Idempotency] = None
        metadata: Optional[resemble.aio.types.GrpcMetadata] = None
        bearer_token: Optional[str] = None
        if __options__ is not None:
            if __options__.idempotency_key is not None:
                idempotency = Idempotency(key=__options__.idempotency_key)
            elif __options__.idempotency_alias is not None:
                idempotency = Idempotency(alias=__options__.idempotency_alias)
            if __options__.metadata is not None:
                metadata = __options__.metadata
            if __options__.bearer_token is not None:
                bearer_token = __options__.bearer_token

        if isinstance(__context_or_workflow__, WorkflowContext) and idempotency is None:
            raise ValueError(
                'Calls to mutators within a `workflow` must use idempotency'
            )

        reference = cls.lookup(__actor_id__, bearer_token=bearer_token)
        stub = reference._writer(__context_or_workflow__)
        return (
            reference,
            await stub.Send(
                request,
                idempotency=idempotency,
                metadata=metadata,
                bearer_token=bearer_token,
            )
        )


    class WeakReference:

        def __init__(
            self,
            actor_id: str,
            *,
            bearer_token: Optional[str] = None,
        ):
            resemble.aio.call.validate_ascii(
                actor_id,
                'actor_id',
                MAX_ACTOR_ID_LENGTH,
                error_type=InvalidActorIdError,
                illegal_characters='\x00\x01',
            )
            self._actor_id = actor_id
            self._reader_stub: Optional[MessageReaderStub] = None
            self._writer_stub: Optional[MessageWriterStub] = None
            self._workflow_stub: Optional[MessageWorkflowStub] = None
            self._tasks_stub: Optional[MessageTasksStub] = None
            self._bearer_token = bearer_token

        def _reader(
            self,
            context_or_workflow: ReaderContext | WriterContext | TransactionContext | WorkflowContext | Workflow,
        ) -> MessageReaderStub:
            if self._reader_stub is None:
                self._reader_stub = MessageReaderStub(
                    context_or_workflow=context_or_workflow,
                    actor_id=self._actor_id,
                    bearer_token=self._bearer_token,
                )
            assert self._reader_stub is not None
            assert self._reader_stub._idempotency_manager == context_or_workflow
            return self._reader_stub

        def _writer(
            self,
            context_or_workflow: TransactionContext | WorkflowContext | Workflow,
        ) -> MessageWriterStub:
            if self._writer_stub is None:
                self._writer_stub = MessageWriterStub(
                    context_or_workflow=context_or_workflow,
                    actor_id=self._actor_id,
                    bearer_token=self._bearer_token,
                )
            assert self._writer_stub is not None
            if self._writer_stub._idempotency_manager != context_or_workflow:
                raise MixedContextsError(
                    "This `WeakReference` for `Message` with ID "
                    f"'{self._actor_id}' has previously been used by a "
                    "different `Context` or `Workflow`. That is not allowed. "
                    "Instead create a new `WeakReference` for every `Context` "
                    "and `Workflow`, by calling "
                    f"`Message.lookup('{self._actor_id}')`."
                )
            return self._writer_stub

        def _workflow(
            self,
            context_or_workflow: WorkflowContext | Workflow,
        ) -> MessageWorkflowStub:
            if self._workflow_stub is None:
                self._workflow_stub = MessageWorkflowStub(
                    context_or_workflow=context_or_workflow,
                    actor_id=self._actor_id,
                    bearer_token=self._bearer_token,
                )
            assert self._workflow_stub is not None
            assert self._workflow_stub._idempotency_manager == context_or_workflow
            return self._workflow_stub

        def _tasks(
            self,
            context_or_workflow: TransactionContext | WorkflowContext | Workflow,
        ) -> MessageTasksStub:
            if self._tasks_stub is None:
                self._tasks_stub = MessageTasksStub(
                    context_or_workflow=context_or_workflow,
                    actor_id=self._actor_id,
                    bearer_token=self._bearer_token,
                )
            assert self._tasks_stub is not None
            assert self._tasks_stub._idempotency_manager == context_or_workflow
            return self._tasks_stub

        class Reactively:

            def __init__(
                self,
                *,
                service_name: ServiceName,
                actor_id: ActorId,
                bearer_token: Optional[str] = None,
            ):
                self._service_name = service_name
                self._actor_id = actor_id
                self._bearer_token = bearer_token


        def reactively(self):
            return Message.WeakReference.Reactively(
                service_name='resemble.integrations.mailgun.v1.Message',
                actor_id=self._actor_id,
                bearer_token=self._bearer_token,
            )

        def schedule(
            self,
            *,
            when: Optional[datetime | timedelta] = None,
        ) -> Any: # TODO: Avoid the nesting as it breaks mypy.
            class Tasks:
                def __init__(
                    self,
                    tasks: Callable[[TransactionContext | WorkflowContext | Workflow], MessageTasksStub],
                    *,
                    when: Optional[datetime | timedelta] = None,
                ) -> None:
                    self._tasks = tasks
                    self._when = when

                # Message callable tasks:
                async def SendTask(
                    self,
                    __context_or_workflow__: TransactionContext | WorkflowContext | Workflow,
                    __options__: Optional[resemble.aio.call.Options] = None,
                    *,
                    send_request: Optional[resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest] = None,
                ) -> Message.SendTaskFuture:
                    if send_request is not None and not isinstance(
                        send_request,
                        resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest,
                    ):
                        raise TypeError(
                            f"Can not construct protobuf message of type "
                            f"'resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest': field 'send_request' is not "
                            f"of required type 'resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest'"
                        )
                    # TODO: mypy-protobuf declares that
                    # `google.protobuf.message.Message` constructor arguments are
                    # always non-None, when in reality they are optional.
                    request = resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest(
                        send_request=send_request,  # type: ignore[arg-type]
                    )

                    schedule: Optional[datetime] = (datetime.now() + self._when) if isinstance(
                        self._when, timedelta
                    ) else self._when

                    metadata: Optional[resemble.aio.types.GrpcMetadata] = None
                    idempotency: Optional[resemble.aio.idempotency.Idempotency] = None
                    bearer_token: Optional[str] = None
                    if __options__ is not None:
                        if __options__.idempotency_key is not None:
                            idempotency = Idempotency(key=__options__.idempotency_key)
                        elif __options__.idempotency_alias is not None:
                            idempotency = Idempotency(alias=__options__.idempotency_alias)
                        if __options__.metadata is not None:
                            metadata = __options__.metadata
                        if __options__.bearer_token is not None:
                            bearer_token = __options__.bearer_token

                    # Add scheduling information to the metadata.
                    metadata = (
                        (resemble.aio.headers.TASK_SCHEDULE,
                        schedule.isoformat() if schedule else ''),
                    ) + (metadata or tuple())

                    if isinstance(__context_or_workflow__, WorkflowContext) and idempotency is None:
                        raise ValueError(
                            'Calls to mutators within a `workflow` must use idempotency'
                        )

                    return Message.SendTaskFuture(
                        __context_or_workflow__,
                        task_id=await self._tasks(__context_or_workflow__).SendTask(
                            request,
                            idempotency=idempotency,
                            metadata=metadata,
                            bearer_token=bearer_token,
                        )
                    )

            return Tasks(self._tasks, when=when)

        # Message specific methods:
        async def SendTask(
            self,
            __context_or_workflow__: TransactionContext | WorkflowContext | Workflow,
            __options__: Optional[resemble.aio.call.Options] = None,
            *,
            send_request: Optional[resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest] = None,
        ) -> resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse:
            if send_request is not None and not isinstance(
                send_request,
                resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest,
            ):
                raise TypeError(
                    f"Can not construct protobuf message of type "
                    f"'resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest': field 'send_request' is not "
                    f"of required type 'resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest'"
                )
            # TODO: mypy-protobuf declares that
            # `google.protobuf.message.Message` constructor arguments are
            # always non-None, when in reality they are optional.
            request = resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest(
                send_request=send_request,  # type: ignore[arg-type]
            )
            idempotency: Optional[resemble.aio.idempotency.Idempotency] = None
            metadata: Optional[resemble.aio.types.GrpcMetadata] = None
            bearer_token: Optional[str] = None
            if __options__ is not None:
                if __options__.idempotency_key is not None:
                    idempotency = Idempotency(key=__options__.idempotency_key)
                elif __options__.idempotency_alias is not None:
                    idempotency = Idempotency(alias=__options__.idempotency_alias)
                if __options__.metadata is not None:
                    metadata = __options__.metadata
                if __options__.bearer_token is not None:
                    bearer_token = __options__.bearer_token

            if isinstance(__context_or_workflow__, WorkflowContext) and idempotency is None:
                raise ValueError(
                    'Calls to mutators within a `workflow` must use idempotency'
                )

            return await self._writer(__context_or_workflow__).SendTask(
                request,
                idempotency=idempotency,
                metadata=metadata,
                bearer_token=bearer_token,
            )

    @classmethod
    def lookup(
        cls,
        actor_id: str,
        *,
        bearer_token: Optional[str] = None,
    ) -> Message.WeakReference:
        return Message.WeakReference(
            actor_id, bearer_token=bearer_token
        )


class MessageServicerNodeAdaptor(Message.Interface):

    def __init__(self):
        # TODO: See https://github.com/reboot-dev/respect/issues/2339
        self._js_servicer_reference = self._construct_js_servicer()  # type: ignore[attr-defined]

    # Message specific methods:
    async def Send(
        self,
        context: WriterContext,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendRequest,
    ) -> Message.SendEffects:

        json_request = google.protobuf.json_format.MessageToJson(request)

        # TODO: See https://github.com/reboot-dev/respect/issues/2339
        json_effects = await self._writer(  # type: ignore[attr-defined]
            self._js_servicer_reference,
            'Send',
            context,
            None,
            json_request,
        )

        effects = json.loads(json_effects)

        assert not hasattr(effects, 'tasks'), "JS tasks not yet implemented"

        return Message.SendEffects(
            state=google.protobuf.json_format.ParseDict(
                effects['state'],
                resemble.integrations.mailgun.v1.mailgun_pb2.MessageState(),
            ),
            response=google.protobuf.json_format.ParseDict(
                effects['response'],
                resemble.integrations.mailgun.v1.mailgun_pb2.SendResponse(),
            ),
        )

    async def SendTask(
        self,
        context: WriterContext,
        state: resemble.integrations.mailgun.v1.mailgun_pb2.MessageState,
        request: resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskRequest,
    ) -> Message.SendTaskEffects:
        json_state = google.protobuf.json_format.MessageToJson(state)

        json_request = google.protobuf.json_format.MessageToJson(request)

        # TODO: See https://github.com/reboot-dev/respect/issues/2339
        json_effects = await self._writer(  # type: ignore[attr-defined]
            self._js_servicer_reference,
            'SendTask',
            context,
            json_state,
            json_request,
        )

        effects = json.loads(json_effects)

        assert not hasattr(effects, 'tasks'), "JS tasks not yet implemented"

        return Message.SendTaskEffects(
            state=google.protobuf.json_format.ParseDict(
                effects['state'],
                resemble.integrations.mailgun.v1.mailgun_pb2.MessageState(),
            ),
            response=google.protobuf.json_format.ParseDict(
                effects['response'],
                resemble.integrations.mailgun.v1.mailgun_pb2.SendTaskResponse(),
            ),
        )



class MessageNodeAdaptor(Message.WeakReference):

    async def _reader(  # type: ignore[override]
        self,
        method: str,
        context: ReaderContext | WriterContext | TransactionContext | WorkflowContext | Workflow,
        request_type: type[google.protobuf.message.Message],
        json_request: str,
    ) -> str:
        request = request_type()

        google.protobuf.json_format.Parse(json_request, request)

        response = await getattr(super()._reader(context), method)(request)

        return google.protobuf.json_format.MessageToJson(response)

    async def _writer(  # type: ignore[override]
        self,
        method: str,
        context: TransactionContext | WorkflowContext | Workflow,
        request_type: type[google.protobuf.message.Message],
        json_request: str,
    ) -> str:
        request = request_type()

        google.protobuf.json_format.Parse(json_request, request)

        response = await getattr(super()._writer(context), method)(request)

        return google.protobuf.json_format.MessageToJson(response)

# yapf: enable
